@@ -1,11 +1,45 @@
+/*
+ * lwm, a window manager for X11
+ * Copyright (C) 1997-2003 Elliott Hughes, James Carter
+ * 
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License
+ * as published by the Free Software Foundation; either version 2
+ * of the License, or (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
+ */
+
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 
 #include <X11/X.h>
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/Xatom.h>
 
+/* These are Motif definitions from Xm/MwmUtil.h, but Motif isn't available
+   everywhere. */
+#define MWM_HINTS_FUNCTIONS     (1L << 0)
+#define MWM_HINTS_DECORATIONS   (1L << 1)
+#define MWM_HINTS_INPUT_MODE    (1L << 2)
+#define MWM_HINTS_STATUS        (1L << 3)
+#define MWM_DECOR_ALL           (1L << 0)
+#define MWM_DECOR_BORDER        (1L << 1)
+#define MWM_DECOR_RESIZEH       (1L << 2)
+#define MWM_DECOR_TITLE         (1L << 3)
+#define MWM_DECOR_MENU          (1L << 4)
+#define MWM_DECOR_MINIMIZE      (1L << 5)
+#define MWM_DECOR_MAXIMIZE      (1L << 6)
+
 #include "lwm.h"
 
 static int getProperty(Window, Atom, Atom, long, unsigned char **);
@@ -30,6 +64,27 @@ manage(Client * c, int mapped)
 	static int	auto_x = 100;
 	static int	auto_y = 100;
 
+	/* get the EWMH window type, as this might overrule some hints */
+	c->wtype = ewmh_get_window_type(c->window);
+	/* get in the initial EWMH state */
+	ewmh_get_state(c);
+	/* set EWMH allowable actions, now we intend to manage this window */
+	ewmh_set_allowed(c);
+	/* is this window to have a frame? */
+	if (c->wtype == WTypeNone) {
+		/* this breaks the ewmh spec (section 5.6) because in the
+		 * absence of a _NET_WM_WINDOW_TYPE, _WM_WINDOW_TYPE_NORMAL
+		 * must be taken. bummer.
+		 */
+		c->framed = motifWouldDecorate(c);
+	} else {
+		c->framed = ewmh_hasframe(c);
+	}
+	if (isShaped(c->window)) c->framed = False;
+
+	/* get the EWMH strut - if there is one */
+	ewmh_get_strut(c);
+
 	/*
 	 * Get the hints, window name, and normal hints (see ICCCM
 	 * section 4.1.2.3).
@@ -74,6 +129,9 @@ manage(Client * c, int mapped)
 	/* Work out details for the Client structure from the hints. */
 	if (hints && (hints->flags & InputHint))
 		c->accepts_focus = hints->input;
+	if (c->proto | Ptakefocus)
+		/* WM_TAKE_FOCUS overrides normal hints */
+		c->accepts_focus = True;
 
 	if (!getWindowState(c->window, &state))
 		state = hints ? hints->initial_state : NormalState;
@@ -96,8 +154,12 @@ manage(Client * c, int mapped)
 		 * "The size specifiers refer to the width and height of the
 		 * client excluding borders" -- ICCCM 4.1.2.3.
 		 */
-		c->size.width  = w + 2 * border;
-		c->size.height = h + 2 * border;
+		c->size.width  = w;
+		c->size.height = h;
+		if (c->framed == True) {
+			c->size.width  += 2 * border;
+			c->size.height += 2 * border;
+		}
 
 		/*
 		 * THIS IS A HACK!
@@ -115,13 +177,16 @@ manage(Client * c, int mapped)
 		/*
 		 * If we have a user-specified position for a top-level window,
 		 * or a program-specified position for a dialogue box, we'll
-		 * take it.
+		 * take it. We'll also just take it during initialisation,
+		 * since the previous manage probably placed its windows
+		 * sensibly.
 		 */
 		if (c->trans != None && c->size.flags & PPosition) {
 			/* It's a "dialogue box". Trust it. */
 			c->size.x = x;
 			c->size.y = y;
-		} else if (c->size.flags & USPosition) {
+		} else if ((c->size.flags & USPosition) ||
+			c->framed == False || mode == wm_initialising ) {
 			/* Use the specified window position. */
 			c->size.x = x;
 			c->size.y = y;
@@ -136,34 +201,57 @@ manage(Client * c, int mapped)
 			/* No position was specified: use the auto-placement
 			 * heuristics. */
 
-			if (auto_x + c->size.width > c->screen->display_width && c->size.width <= c->screen->display_width) {
+			/* firstly, make sure auto_x and auto_y are outside
+			 * strut */
+			if (auto_x < c->screen->strut.left)
+				auto_x = c->screen->strut.left;
+			if (auto_y < c->screen->strut.top)
+				auto_y = c->screen->strut.top;
+
+			if ((auto_x + c->size.width) > 
+				(c->screen->display_width -
+				c->screen->strut.right) && 
+				(c->size.width <=
+				(c->screen->display_width -
+				c->screen->strut.left -
+				c->screen->strut.right))) {
 				/*
-				 * If the window wouldn't fit using normal auto-placement but is
-				 * small enough to fit horizontally, then centre the window
+				 * If the window wouldn't fit using normal
+				 * auto-placement but is small enough to fit
+				 * horizontally, then centre the window
 				 * horizontally.
 				 */
-				c->size.x = (c->screen->display_width - c->size.width) / 2;
-				auto_x = 20;
+				c->size.x = (c->screen->display_width
+					- c->size.width) / 2;
+				auto_x = c->screen->strut.left + 20;
 			} else {
 				c->size.x = auto_x;
 				auto_x += 10;
-				if (auto_x > 500)
-					auto_x = 20;
+				if (auto_x > (c->screen->display_width / 2))
+					auto_x = c->screen->strut.left + 20;
 			}
 
-			if (auto_y + c->size.height > c->screen->display_height && c->size.height <= c->screen->display_height) {
+			if (((auto_y + c->size.height) >
+				(c->screen->display_height -
+				c->screen->strut.bottom)) &&
+				(c->size.height <=
+				(c->screen->display_height -
+				c->screen->strut.top -
+				c->screen->strut.bottom))) {
 				/*
-				 * If the window wouldn't fit using normal auto-placement but is
-				 * small enough to fit vertically, then centre the window
+				 * If the window wouldn't fit using normal
+				 * auto-placement but is small enough to fit
+				 * vertically, then centre the window
 				 * vertically.
 				 */
-				 c->size.y = (c->screen->display_height - c->size.height) / 2;
-				 auto_y = 20;
+				 c->size.y = (c->screen->display_height
+					- c->size.height) / 2;
+				 auto_y = c->screen->strut.top + 20;
 			} else {
 				c->size.y = auto_y;
 				auto_y += 10;
-				if (auto_y > 500)
-					auto_y = 20;
+				if (auto_y > (c->screen->display_height / 2))
+					auto_y = c->screen->strut.top + 20;
 			}
 		}
 	}
@@ -175,14 +263,15 @@ manage(Client * c, int mapped)
 	 * Do all the reparenting and stuff.
 	 */
 
-	if (!isShaped(c->window)) {
+	if (c->framed == True) {
 		c->parent = XCreateSimpleWindow(dpy, c->screen->root,
 			c->size.x, c->size.y - titleHeight(),
 			c->size.width, c->size.height + titleHeight(),
 			1, c->screen->black, c->screen->white);
 
 		attr.event_mask = ExposureMask | EnterWindowMask | ButtonMask |
-			SubstructureRedirectMask | SubstructureNotifyMask;
+			SubstructureRedirectMask | SubstructureNotifyMask |
+			PointerMotionMask;
 		XChangeWindowAttributes(dpy, c->parent, CWEventMask, &attr);
 
 		XResizeWindow(dpy, c->window, c->size.width - 2 * border,
@@ -200,14 +289,19 @@ manage(Client * c, int mapped)
 		XSetWindowBorderWidth(dpy, c->window, 0);
 
 	attr.event_mask = ColormapChangeMask | EnterWindowMask |
-		PropertyChangeMask;
+		PropertyChangeMask | FocusChangeMask;
 	attr.win_gravity = StaticGravity;
 	attr.do_not_propagate_mask = ButtonMask;
 	XChangeWindowAttributes(dpy, c->window,
 		CWEventMask | CWWinGravity | CWDontPropagate, &attr);
 
-	XReparentWindow(dpy, c->window, c->parent,
-		border, border + titleHeight());
+	if (c->framed == True) {
+		XReparentWindow(dpy, c->window, c->parent,
+			border, border + titleHeight());
+	} else {
+		XReparentWindow(dpy, c->window, c->parent,
+			c->size.x, c->size.y);
+	}
 
 	setShape(c);
 
@@ -219,10 +313,12 @@ manage(Client * c, int mapped)
 		c->hidden = False;
 		XMapWindow(dpy, c->parent);
 		XMapWindow(dpy, c->window);
-		setactive(c, 0, 0L);
+		setactive(c, (focus_mode == focus_click) ? 1 : 0, 0L);
 		Client_SetState(c, NormalState);
 	}
 
+	if (c->wstate.fullscreen == True) Client_EnterFullScreen(c);
+
 	if (current != c)
 		cmapfocus(current);
 }
@@ -349,7 +445,14 @@ Terminate(int signal) {
 	installColourmap(None);
 	
 	XCloseDisplay(dpy);
-	exit(EXIT_FAILURE);
+
+	session_end();
+
+	if (signal) {
+		exit(EXIT_FAILURE);
+	} else {
+		exit(EXIT_SUCCESS);
+	}
 }
 
 
@@ -378,7 +481,6 @@ getProperty(Window w, Atom a, Atom type, long len, unsigned char **p) {
 void
 getWindowName(Client *c) {
 	char * name;
-	char * machine;
 	Atom actual_type;
 	int format;
 	unsigned long n;
@@ -390,44 +492,42 @@ getWindowName(Client *c) {
 	
 	was_nameless = (c->name == 0);
 	
-	if (XGetWindowProperty(dpy, c->window, _mozilla_url, 0L, 100L, False, AnyPropertyType, &actual_type, &format, &n, &extra, (unsigned char **) &name) == Success && name && *name != '\0' && n != 0) {
+	if (ewmh_get_window_name(c) == True) {
+		c->name_utf8 = True;
+	} else if (XGetWindowProperty(dpy, c->window, _mozilla_url, 0L, 100L, False, AnyPropertyType, &actual_type, &format, &n, &extra, (unsigned char **) &name) == Success && name && *name != '\0' && n != 0) {
 		if (c->name != 0)
 			XFree(c->name);
 		c->name = name;
+		c->namelen = n;
+		c->name_utf8 = False;
 	} else if (XGetWindowProperty(dpy, c->window, XA_WM_NAME, 0L, 100L, False, AnyPropertyType, &actual_type, &format, &n, &extra, (unsigned char **) &name) == Success && name && *name != '\0' && n != 0) {
 		/* That rather unpleasant condition is necessary because xwsh uses
-		 * COMPOUND_TEXT rather than STRING for its WM_NAME property,
-		 * and anonymous xwsh windows are annoying.
-		 */
+	 	* COMPOUND_TEXT rather than STRING for its WM_NAME property,
+	 	* and anonymous xwsh windows are annoying.
+	 	*/
 		if (c->name != 0)
 			XFree(c->name);
 		if (actual_type == compound_text && memcmp(name, "\x1b\x28\x42", 3) == 0) {
+			/* FIXME: this breaks i18n, use memcpy? */
 			c->name = sdup(name + 3);
+			c->namelen = strlen(c->name);
+			c->name_utf8 = False;
 			XFree(name);
 		} else {
 			c->name = name;
+			c->namelen = n;
+			c->name_utf8 = False;
 		}
-		
-#ifdef PREPEND_CLIENT_MACHINE
-		/* Prepend the client's machine name, if we know it. */
-		if (c->name && XGetWindowProperty(dpy, c->window, XA_WM_CLIENT_MACHINE, 0L, 100L, False, AnyPropertyType, &actual_type, &format, &n, &extra, (unsigned char **) &machine) == Success && machine && *machine != '\0' && n != 0) {
-			int len = strlen(machine) + strlen(c->name) + 2;
-			char * new_name = (char *) malloc(len);
-			sprintf(new_name, "%s:%s", machine, c->name);
-			XFree(machine);
-			XFree(c->name);
-			c->name = new_name;
-		}
-#endif
 	}
 	
-	if (c->name == 0)
+	if (c->name == 0) {
+		c->namelen = 0;
+		c->name_utf8 = False;
 		return;
+	}
 	
 	if (!was_nameless)
 		Client_DrawBorder(c, c == current);
-	if (hidden(c))
-		update_menu(c);
 }
 
 void
@@ -448,29 +548,31 @@ getNormalHints(Client *c) {
 	if (XGetWMNormalHints(dpy, c->window, &c->size, &msize) == 0)
 		c->size.flags = 0;
 
-	/*
-	 * Correct the minimum allowable size of this client to take account of
-	 * the window border.
-	 */
-	if (c->size.flags & PMinSize) {
-		c->size.min_width  += 2 * border;
-		c->size.min_height += 2 * border;
-	} else {
-		c->size.flags |= PMinSize;
-		c->size.min_width  = 2 * (2 * border);
-		if (c->accepts_focus)
-			c->size.min_height = 2 * (2*border);
-		else
-			c->size.min_height = 2 * (2*border);
-	}
+	if (c->framed == True) {
+		/*
+	 	* Correct the minimum allowable size of this client to take
+		* account of the window border.
+	 	*/
+		if (c->size.flags & PMinSize) {
+			c->size.min_width  += 2 * border;
+			c->size.min_height += 2 * border;
+		} else {
+			c->size.flags |= PMinSize;
+			c->size.min_width  = 2 * (2 * border);
+			if (c->accepts_focus)
+				c->size.min_height = 2 * (2*border);
+			else
+				c->size.min_height = 2 * (2*border);
+		}
 
-	/*
-	 * Correct the maximum allowable size of this client to take account of
-	 * the window border.
-	 */
-	if (c->size.flags & PMaxSize) {
-		c->size.max_width  += 2 * border;
-		c->size.max_height += 2 * border;
+		/*
+	 	* Correct the maximum allowable size of this client to take
+		* account of the window border.
+	 	*/
+		if (c->size.flags & PMaxSize) {
+			c->size.max_width  += 2 * border;
+			c->size.max_height += 2 * border;
+		}
 	}
 
 	/*
@@ -511,3 +613,21 @@ getWindowState(Window w, int *state) {
 	XFree(p);
 	return 1;
 }
+
+
+extern Bool
+motifWouldDecorate(Client *c) {
+	unsigned long *p = 0;
+	Bool ret = True; /* if all else fails - decorate */
+
+	if (getProperty(c->window, motif_wm_hints, motif_wm_hints,
+		5L, (unsigned char **) &p) <= 0) 
+		return ret;
+
+	if ((p[0] & MWM_HINTS_DECORATIONS) &&
+		!(p[2] & (MWM_DECOR_BORDER | MWM_DECOR_ALL)))
+		ret = False;
+
+	XFree(p);
+	return ret;
+}
